import pathlib
import typing
from unittest.mock import patch

import lobotomy as lbm


class Patch:
    """
    A patch decorator class that will behave like the unittest.mock.patch
    specifically for the boto3.Session object, replacing it in the test
    with the Lobotomy session factory that will be used to replace the normal
    boto3 calls with the lobotomized ones.
    """

    def __init__(
        self,
        path: typing.Union[None, str, pathlib.Path] = None,
        data: dict = None,
        prefix: typing.Union[None, str, typing.Iterable[str]] = None,
        patch_path: str = "boto3.Session",
        client_overrides: typing.Dict[str, typing.Any] = None,
    ):
        """
        Create a patch instance.

        :param path:
            Path to a configuration file containing the lobotomy data to use
            in mock calls during the lifetime of this patch. This can be the
            default value if a data value is set instead.
        :param data:
            Dictionary containing the lobotomy data to use in mock calls
            during the lifetime of this patch. This can be the default None
            value if a path value is set instead.
        :param prefix:
            An optional key or multi-key prefix within the path-loaded data
            object where the lobotomy data resides. Use this when the
            configuration file for the test has a broader scope.
        :param patch_path:
            The targeted patch target, which defaults to 'boto3.Session'.
            This can be used to override that value for cases where a
            different patch target is desirable.
        :param client_overrides:
            Optional dictionary containing mappings of service names to clients
            that should be used in place of lobotomy clients during the lifecycle
            of this lobotomy patch. Useful for mixing lobotomy clients with other
            mocking clients in complex testing scenarios.
        """
        self.data = data
        self.path = path
        self.prefix = prefix
        self.patch_path = patch_path
        self.client_overrides = client_overrides

    def _make_lobotomy(self) -> "lbm.Lobotomy":
        """
        Creates the lobotomy object to be used during the patch lifetime.
        This has to be created with each call to prevent multiple scenario
        executions for the same test function.
        """
        if self.data is not None:
            return lbm.Lobotomy(
                self.data.copy(),
                client_overrides=self.client_overrides,
            )
        elif self.path:
            return lbm.Lobotomy.from_file(
                self.path,
                prefix=self.prefix,
                client_overrides=self.client_overrides,
            )

        return lbm.Lobotomy(client_overrides=self.client_overrides)

    def __call__(self, caller: typing.Callable):
        """
        Calls the patching decorator that wraps the test function with
        a patch of the boto3.Session class.
        """
        return patch(self.patch_path, new_callable=self._make_lobotomy)(caller)



import functools
import typing
from unittest.mock import MagicMock

import lobotomy
from lobotomy import _services
from lobotomy._clients import _casting
from lobotomy._clients import _validation


class Client:
    """
    Mocks AWS boto3.client behaviors that pull response data from data stored
    in configuration files.
    """

    def __init__(
        self,
        session: "lobotomy.Session",
        service_name: str,
        *args,
        **kwargs,
    ):
        """Populate with the loaded scenario data."""
        self._service_name = service_name
        self._client_init_args: tuple = args
        self._client_init_kwargs: dict = kwargs
        self._calls: typing.List[lobotomy.ServiceCall] = []
        self._service: _services.Service = _services.load_definition(
            service_name,
        )
        self._session = session

        self.exceptions = MagicMock()
        self._registered_exceptions = {}
        for name in self._service.exceptions:
            new_type = type(name, (lobotomy.ClientError,), {})
            setattr(self.exceptions, name, new_type)
            self._registered_exceptions[name] = new_type

    def _call(self, called_method_name: str, *args, **kwargs):
        raw = self._session.lobotomy.pop_response(
            self._service_name,
            called_method_name,
        )
        method = self._service.lookup(called_method_name)
        request = _validation.validate_input(method, args, kwargs)
        response = _casting.cast(method.output, raw)

        service_call = lobotomy.ServiceCall(
            service=self._service_name,
            method=called_method_name,
            request=request,
            args=args,
            kwargs=kwargs,
            response=response,
        )

        self._session.lobotomy.record_call(service_call)
        self._calls.append(service_call)

        if "Error" in response:
            error = self._registered_exceptions.get(
                response["Error"]["Code"],
                lobotomy.ClientError,
            )()
            raise error.populate(**response["Error"])
        return response

    def __getattr__(self, item: str):
        """
        Retrieves response data for the given client call from the
        scenario data that defines the execution.
        """
        if not self._service.has(item):
            raise lobotomy.NoSuchMethod(
                f"""
                No boto/botocore definition found for "{self._service_name}{item}()".
                The version of botocore installed could be out of date, or newer
                and this service method has been removed. Please check the boto3
                documentation to confirm the existence of this service method.
                """
            )

        return functools.partial(self._call, item)

    def get_paginator(self, item: str) -> MagicMock:
        """Mocks a single-page paginator response."""

        def _call(*args, **kwargs):
            return [self._call(item, *args, **kwargs)]

        paginator = MagicMock()
        paginator.paginate.side_effect = _call
        return paginator



import typing
import argparse


def parse(arguments: typing.List[str] = None) -> argparse.Namespace:
    """Parses command line arguments for command execution."""
    parser = argparse.ArgumentParser(
        prog="lobotomy",
        description="""
            Command line interface for lobotomy configuration
            management.
            """,
    )
    subparsers = parser.add_subparsers(title="command", dest="command")

    add_parser = subparsers.add_parser("add")
    add_parser.add_argument("boto_operation")
    add_parser.add_argument("configuration_file_path")
    add_parser.add_argument(
        "--format",
        "--file-format",
        dest="file_format",
        choices=["yaml", "toml", "json"],
    )
    add_parser.add_argument("--prefix")

    return parser.parse_args(args=arguments)



import json
import pathlib
import typing

import toml
import yaml

from lobotomy import _fio
from lobotomy import _mutator
from lobotomy import _services
from lobotomy._cli import _definitions


def _get_path(
    context: "_definitions.CliContext",
) -> typing.Optional[pathlib.Path]:
    if context.args.configuration_file_path in [None, "-"]:
        return None

    return pathlib.Path(context.args.configuration_file_path).expanduser().absolute()


def run(context: "_definitions.CliContext") -> "_definitions.ExecutionResult":
    """Adds a new call to the specified command actions."""
    file_format = context.args.file_format
    path = _get_path(context)
    prefix = [item for item in (context.args.prefix or "").split(".") if item]

    try:
        configs = _fio.read(path, prefix, file_format=file_format)  # type: ignore
    except (AttributeError, FileNotFoundError, TypeError):
        configs = {}

    service_name, method_name = context.args.boto_operation.split(".")
    service = _services.load_definition(service_name)
    method = service.lookup(method_name)

    _mutator.add_service_response(configs, method)

    if path:
        _fio.write(path, configs, prefix, file_format=file_format)
        return _definitions.ExecutionResult(
            code="ADDED", message="New call has been added to the configs."
        ).echo()

    if file_format == "json":
        print(json.dumps(configs, indent=2))
    elif file_format == "toml":
        print(toml.dumps(configs))
    else:
        print(yaml.safe_dump(configs))

    return _definitions.ExecutionResult(
        code="ECHOED", message="New call has been echoed to stdout."
    )



import json
import pathlib
import typing

import toml
import yaml


def _read_file(path: pathlib.Path, file_format: str = None) -> dict:
    """Reads the configuration file from disk if it exists."""
    p = pathlib.Path(path).expanduser().absolute()
    if not p.exists():
        return {}

    contents = p.read_text()

    if file_format == "yaml" or path.name.endswith((".yaml", ".yml")):
        return yaml.safe_load(contents)

    if file_format == "toml" or path.name.endswith(".toml"):
        return typing.cast(dict, toml.loads(contents))

    return json.loads(contents)


def read(
    path: typing.Union[str, pathlib.Path],
    prefix: typing.Union[str, typing.Iterable[str]] = None,
    file_format: str = None,
) -> typing.Dict[str, typing.Any]:
    """
    Reads the specified configuration file and returns the result. If a
    prefix is specified, the data at the given prefix will be returned
    instead.

    :param path:
        Path to the configuration file to read.
    :param prefix:
        Optional prefix within the file in which the configuration resides.
    :param file_format:
        Optional format for the file, which will be determined from the
        file extension if omitted.
    :return:
        A dictionary containing the lobotomy configuration data from the
        given file. This will be an empty dictionary if no configuration
        exists yet.
    """
    data = _read_file(pathlib.Path(path), file_format)
    if not prefix:
        return data or {}

    keys = [prefix] if isinstance(prefix, str) else prefix
    output = data
    for key in keys:
        output = output.get(key) or {}
    return output


def write(
    path: typing.Union[str, pathlib.Path],
    configs: typing.Dict[str, typing.Any],
    prefix: typing.Union[str, typing.Iterable[str]] = None,
    file_format: str = None,
) -> None:
    """
    Writes the specified configuration file with updated configuration data.
    If a prefix is specified, the data will be inserted at the given prefix.

    :param path:
        Path to the configuration file to read.
    :param configs:
        Updated lobotomy configuration data to be written to the file.
        This will replace the existing configuration data in the file.
    :param prefix:
        Optional prefix within the file in which the configuration resides.
    :param file_format:
        Optional format for the file, which will be determined from the
        file extension if omitted.
    :return:
        A dictionary containing the lobotomy configuration data from the
        given file. This will be an empty dictionary if no configuration
        exists yet.
    """
    keys = [prefix] if isinstance(prefix, str) else prefix
    path = pathlib.Path(path).expanduser().absolute()
    data = _read_file(path, file_format)

    child = data
    for key in keys or []:
        if key not in child:
            child[key] = {}
        child = child[key]

    child.update(configs)

    if file_format == "yaml" or path.name.endswith((".yaml", ".yml")):
        contents = yaml.safe_dump(data)
    elif file_format == "toml" or path.name.endswith(".toml"):
        contents = toml.dumps(data)
    else:
        contents = json.dumps(data, indent=2)

    path.write_text(contents)



import datetime
import typing
from unittest.mock import MagicMock

import dateutil.parser
from botocore.response import StreamingBody

import lobotomy


class InternalStreamer:
    """
    An internal class that replaces the HTTPResponse object that boto
    StreamingBody objects wrap. This is used to simulate the streaming
    body behaviors in returned streaming response objects.
    """

    def __init__(self, source: typing.Union[bytes, str]):
        """Wrap the source string in a streaming interface."""
        self._cursor = 0
        self._source: typing.Union[bytes, str] = source

        # This exists for botocore compatibility setting timeouts.
        self._fp = MagicMock()

    def read(self, amt: int = None):
        """Simulates the streaming interface."""
        length = len(self._source)
        amount = amt if amt is not None else length
        start = max(0, self._cursor)
        end = min(length, self._cursor + amount)
        self._cursor = end
        return self._source[start:end]

    def close(self):
        # Exists only for compatibility cases.
        pass

    def __len__(self):
        return len(self._source)


def _cast_structure(definition: dict, value: dict) -> dict:
    """
    Converts a dictionary/structure botocore type into its formatted values
    by recursively casting all the members.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        The cast version of the specified value that matches the format of
        the value as it would be returned in a boto client response.
    """
    return {k: cast(definition["members"].get(k), v) for k, v in value.items()}


def _cast_noop(definition: dict, value: typing.Any) -> typing.Any:
    """
    A non-casting, no-op operation used for unknown or uncastable value
    types.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        An unmodified value as this is a no-op.
    """
    return value


def _cast_integer(definition: dict, value: typing.Any) -> int:
    """
    Converts a value into an integer value.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        The cast version of the specified value that matches the format of
        the value as it would be returned in a boto client response.
    """
    return int(value)


def _cast_list(definition: dict, value: list) -> list:
    """
    Converts a list botocore type into its formatted values
    by recursively casting all the items.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        The cast version of the specified value that matches the format of
        the value as it would be returned in a boto client response.
    """
    return [cast(definition.get("member"), v) for v in value]


def _cast_string(
    definition: dict,
    value: typing.Any,
) -> typing.Union[str, StreamingBody]:
    """
    Converts a string botocore type into its formatted value. If the
    definition indicates that the string is returned in streaming format,
    a StreamingBody object is returned instead with the string value
    injected into the body via an InternalStreamer object that mimics the
    actual streaming behavior of a real StreamingBody response.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        The cast version of the specified value that matches the format of
        the value as it would be returned in a boto client response.
    """
    output = str(value)
    if definition.get("streaming"):
        output = StreamingBody(InternalStreamer(output), len(output))
    return output


def _cast_blob(
    definition: dict,
    value: typing.Any,
) -> typing.Union[bytes, StreamingBody]:
    """
    Converts a blob botocore type into its formatted value. If the
    definition indicates that the blob is returned in streaming format,
    a StreamingBody object is returned instead with the bytes value
    injected into the body via an InternalStreamer object that mimics the
    actual streaming behavior of a real StreamingBody response.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        The cast version of the specified value that matches the format of
        the value as it would be returned in a boto client response.
    """
    if isinstance(value, bytes):
        output = value
    else:
        output = str(value).encode()

    if definition.get("streaming"):
        output = StreamingBody(InternalStreamer(output), len(output))
    return output


def _cast_timestamp(
    definition: dict,
    value: typing.Union[float, int, str, datetime.date, datetime.datetime],
) -> datetime.datetime:
    """
    Converts a string date definition into a datetime as would be returned
    by the boto response. The dateutil library is used to parse the string
    for flexible representation. Additionally, an integer/float value can
    be specified, which will be treated as a unix timestamp. Note that
    boto returns timezone-aware datetime objects, so this loading process
    ensures that the returned datetime is timezone-aware. If no timezone
    was determined by the loading process, UTC is used by default.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        The cast version of the specified value that matches the format of
        the value as it would be returned in a boto client response.
    """
    if isinstance(value, (int, float)):
        output = datetime.datetime.fromtimestamp(value, tz=datetime.timezone.utc)
    elif isinstance(value, datetime.datetime):
        output = value
    elif isinstance(value, datetime.date):
        output = datetime.datetime(
            year=value.year,
            month=value.month,
            day=value.day,
            tzinfo=datetime.timezone.utc,
        )
    else:
        output = dateutil.parser.parse(value)

    if not output.tzinfo:
        return output.replace(tzinfo=datetime.timezone.utc)
    return output


def cast(definition: typing.Optional[dict], value: typing.Any) -> typing.Any:
    """
    Casts the raw mocked value into its equivalent boto response type as
    defined by the definition object for the value. For complex types, this
    function acts recursively.

    :param definition:
        Specification definition for the associated value to cast.
    :param value:
        A loaded value to be cast into its boto client response value.
    :return:
        The cast version of the specified value that matches the format of
        the value as it would be returned in a boto client response.
    """
    if definition is None:
        # If there is no definition because it does not exist, just return
        # the value as-is.
        return value

    data_type: typing.Optional[str] = definition.get("type")
    conversions = {
        "structure": _cast_structure,
        "list": _cast_list,
        "timestamp": _cast_timestamp,
        "string": _cast_string,
        "long": _cast_integer,
        "integer": _cast_integer,
        "blob": _cast_blob,
        "noop": _cast_noop,
    }
    caster: typing.Any = conversions.get(
        data_type or "noop",
        _cast_noop,
    )

    try:
        return caster(definition, value)
    except Exception as error:
        raise lobotomy.DataTypeError(
            f"""
            Failed to cast specified response data of type "{data_type}"
            for the value:
            {value}
            There may be an incompatibility between the value specified in
            the lobotomy configuration data for the response and the data type
            expected by the return of the boto response.
            """
        ) from error



from lobotomy._services._definitions import Method  # noqa
from lobotomy._services._definitions import Service  # noqa


def load_definition(service_name: str) -> "Service":
    """Loads client definition data for the associated AWS service."""
    return Service(service_name)



import dataclasses
import json
import pathlib
import typing

import toml
import yaml
from botocore.session import Session as BotocoreSession

import lobotomy
from lobotomy import _mutator
from lobotomy import _services


@dataclasses.dataclass(frozen=True)
class ServiceCall:
    """Data structure for recording service calls made on clients."""

    #: Name of the service in which this call was made.
    service: str
    #: Name of the method within the specified service in which this call was made.
    method: str
    #: A normalized request dictionary that contains key/value pairs for each of
    #: the arguments passed into the method call. This combines args and kwargs together
    #: into a representative "kwargs" dictionary to avoid having to manually lookup
    #: method arguments manually in a combined args/kwargs fashion.
    request: typing.Dict[str, typing.Any]
    #: Positional arguments specified in the service method call. These are also
    #: available by keyword in the "request" attribute of this ServiceCall instance.
    args: typing.Tuple[typing.Any, ...]
    #: Keyword arguments specified in the service method call. These are also
    #: available in the "request" attribute of this ServiceCall instance as well.
    kwargs: typing.Dict[str, typing.Any]
    #: The lobotomized response that was returned with this invocation as specified by
    #: the loaded lobotomy data.
    response: typing.Union[list, dict, str, None]


class ReadOnlyCredentials(typing.NamedTuple):
    """Data structure that matches ReadOnlyCredentials from botocore library."""

    access_key: str
    secret_key: str
    token: typing.Optional[str]


class Credentials:
    """
    Mock botocore Credentials object used by Session.get_credentials during
    lobotomy tests.
    """

    def __init__(self, data: dict = None):
        """Stores configuration data for access."""
        self._data = data or {}

    @property
    def method(self) -> str:
        """Authentication method by which credentials were loaded."""
        return self._data.get("method", "manual")

    @property
    def access_key(self) -> str:
        """AWS access key for the associated session."""
        return self._data.get("access_key", "A123LOBOTOMY")

    @property
    def secret_key(self) -> str:
        """AWS secret key for the associated session."""
        return self._data.get("secret_key", "lobotomysecretkey")

    @property
    def token(self) -> typing.Optional[str]:
        """Optional AWS access token for the associated session."""
        return self._data.get("token")

    def get_frozen_credentials(self) -> "ReadOnlyCredentials":
        """A named-tuple form of the credentials for the associated session."""
        return ReadOnlyCredentials(self.access_key, self.secret_key, self.token)


class Session:
    """
    Mock boto3.Session object to use in place of the real one during lobotomy
    tests.
    """

    def __init__(
        self,
        aws_access_key_id: str = None,
        aws_secret_access_key: str = None,
        aws_session_token: str = None,
        region_name: str = None,
        botocore_session: BotocoreSession = None,
        profile_name: str = None,
        source_lobotomy: "Lobotomy" = None,
    ):
        self.lobotomy: Lobotomy = source_lobotomy or Lobotomy()

        self._constructed = {
            "aws_access_key_id": aws_access_key_id,
            "aws_secret_access_key": aws_secret_access_key,
            "aws_session_token": aws_session_token,
            "region_name": region_name,
            "botocore_session": botocore_session,
            "profile_name": profile_name,
        }
        self._definition = self.lobotomy.get_session_data()
        self._data = {
            **self._definition,
            **{k: v for k, v in self._constructed.items() if v is not None},
        }
        self._clients: typing.Dict[str, "lobotomy.Client"] = {}

    @property
    def profile_name(self) -> typing.Optional[str]:
        """Optional name of the AWS profile used for the session if set."""
        return self._data.get("profile_name")

    @property
    def region_name(self) -> typing.Optional[str]:
        """Optional explicit AWS region used for the session if set."""
        return self._data.get("region_name")

    @property
    def available_profiles(self) -> typing.List[str]:
        """List of AWS profiles available for use in the session."""
        return self._data.get("available_profiles") or []

    def get_credentials(self) -> "Credentials":
        """Retrieves the credentials associated with the session."""
        return Credentials(self._data.get("credentials") or {})

    def client(self, service_name: str, *args, **kwargs) -> "lobotomy.Client":
        """
        Creates a lobotomy client object for the specified service. Each
        service-client is a singleton such that subsequent calls for the
        same service name will return the same object.
        """
        override = self.lobotomy.get_client_override(service_name)
        if override is not None:
            return override

        if service_name not in self._clients:
            self._clients[service_name] = lobotomy.Client(
                self,
                service_name,
                *args,
                **kwargs,
            )
        return self._clients[service_name]


class Lobotomy:
    """
    Factory replacement class for boto3.Session that generates mock session
    objects during lobotomy tests.
    """

    def __init__(
        self,
        data: typing.Dict[str, typing.Any] = None,
        client_overrides: typing.Dict[str, typing.Any] = None,
    ):
        self.data = data or {}
        self._client_overrides = client_overrides or {}
        self._service_calls: typing.List[ServiceCall] = []

    @property
    def service_calls(self) -> typing.Tuple["ServiceCall", ...]:
        """A list of the service calls that have been made so far."""
        return tuple(self._service_calls)

    def get_service_call(
        self,
        service_name: str,
        method_name: str,
        index: int = 0,
    ) -> "ServiceCall":
        """
        Finds and returns the Nth service call made for the given method for inspection
        and assertion during test evaluation as specified by the index.

        :param service_name:
            Name of the AWS boto3 service in which the method call being added resides.
        :param method_name:
            Name of the AWS boto3 method to be called for this response within the
            specified service.
        :param index:
            Specifies the Nth service call made for the given service and method names
            to be returned. This is a zero-indexed value.
        :return:
            Nth-specified service call for the given service and method names. If no
            such call exists, an IndexError will be raised.
        """
        return self.get_service_calls(service_name, method_name)[index]

    def get_service_calls(
        self,
        service_name: str,
        method_name: str,
    ) -> typing.List["ServiceCall"]:
        """
        Finds and returns a list containing all service calls made for the given method
        for inspection and assertion during test evaluation.

        :param service_name:
            Name of the AWS boto3 service in which the method call being added resides.
        :param method_name:
            Name of the AWS boto3 method to be called for this response within the
            specified service.
        """
        return [
            s
            for s in self._service_calls
            if s.service == service_name and s.method == method_name
        ]

    def add_client_override(self, service_name: str, client: typing.Any) -> "Lobotomy":
        """
        Add an override client that will be returned instead of a lobotomy client
        within the scope of this Lobotomy's lifecycle.

        :param service_name:
            Name of the AWS boto service of the client to be overridden.
        :param client:
            The replacement client object that will be returned instead of a lobotomy
            client.
        """
        self._client_overrides[service_name] = client
        return self

    def remove_client_override(self, service_name: str) -> "Lobotomy":
        """
        Removes an overridden client if such an override exists. Will do nothing if
        the override does not currently exist.

        :param service_name:
            Name of the AWS boto service of the client to be removed as an override.
        """
        if service_name in self._client_overrides:
            del self._client_overrides[service_name]
        return self

    def get_client_override(self, service_name: str) -> typing.Any:
        """
        Retrieves an override client if it exists.

        :param service_name:
            Name of the AWS boto service for which to fetch an override client.
        """
        return self._client_overrides.get(service_name)

    def add_call(
        self,
        service_name: str,
        method_name: str,
        response: typing.Any = None,
    ) -> "Lobotomy":
        """
        Adds a method call response to the stored data. Operates in the same fashion
        as the CLI, in that additional adds will convert a single call into a list of
        calls.

        :param service_name:
            Name of the AWS boto3 service in which the method call being added resides.
        :param method_name:
            Name of the AWS boto3 method to be called for this response within the
            specified service.
        :param response:
            Boto3 response object for the given method call within the service.
        """
        service = _services.load_definition(service_name)
        method = service.lookup(method_name)
        _mutator.add_service_response(self.data, method, response)
        return self

    def __call__(
        self,
        aws_access_key_id: str = None,
        aws_secret_access_key: str = None,
        aws_session_token: str = None,
        region_name: str = None,
        botocore_session: BotocoreSession = None,
        profile_name: str = None,
    ) -> "Session":
        """
        Mimics the boto3.Session() constructor as a lobotomized session factory.
        The arguments to this method match the arguments to the boto3.Session()
        constructor signature and are stored within the lobotomized session as
        part of the duck-typed interface.
        """
        return Session(
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            region_name=region_name,
            botocore_session=botocore_session,
            profile_name=profile_name,
            source_lobotomy=self,
        )

    def get_session_data(self) -> dict:
        """Retrieves the lobotomized configuration data assocated with the session."""
        data = self.data.get("session", self.data.get("sessions")) or {}
        if not isinstance(data, dict):
            return data.pop(0)
        return data

    def record_call(self, service_call: "ServiceCall") -> None:
        """
        Records a service call made from a client in the list of overall service
        calls for the lobotomized environment. This is meant to be called by
        lobotomy.Clients only.

        :param service_call:
            Service call to record within the lobotomy instance.
        """
        self._service_calls.append(service_call)

    def pop_response(self, service_name: str, method_name: str) -> typing.Any:
        """
        Retrieves the response data for the given service and method name
        combination from the lobotomy data. If such a response does not
        exist, an error is raised instead.

        :param service_name:
            Name of the boto3 service in which the method call lookup is being made.
        :param method_name:
            Name of the method within the boto3 service for which to lookup a response.
        :return:
            A response object containing the lobotomized response for the given
            service method call.
        """
        response = self.data.get("clients", {}).get(service_name, {}).get(method_name)

        if response is None:
            raise lobotomy.NoResponseFound(
                f"""
                No response set for "{service_name}.{method_name}()" in the
                configuration data for this lobotomy. Check to ensure the
                method call was included in the source configuration data.
                """
            )

        if isinstance(response, list):
            if not response:
                raise lobotomy.NoResponseFound(
                    f"""
                    No more responses set for "{service_name}.{method_name}()"
                    in the configuration data for this lobotomy. All specified
                    responses have been returned already.
                    """
                )
            return response.pop(0)
        return response

    @classmethod
    def from_file(
        cls,
        path: typing.Union[str, pathlib.Path],
        prefix: typing.Union[str, typing.Iterable[str]] = None,
        client_overrides: typing.Dict[str, typing.Any] = None,
    ) -> "Lobotomy":
        """
        Create a patch instance.

        :param path:
            Path to a configuration file containing the lobotomy data to use
            in mock calls during the lifetime of this patch. This can be the
            default value if a data value is set instead.
        :param prefix:
            An optional key or multi-key prefix within the path-loaded data
            object where the lobotomy data resides. Use this when the
            configuration file for the test has a broader scope.
        :param client_overrides:
            Optional dictionary containing mappings of service names to clients
            that should be used in place of lobotomy clients during the lifecycle
            of this lobotomy patch. Useful for mixing lobotomy clients with other
            mocking clients in complex testing scenarios.
        """
        p = pathlib.Path(path)
        contents = p.absolute().read_text()

        if p.name.endswith((".yaml", ".yml")):
            data = yaml.safe_load(contents)
        elif p.name.endswith(".toml"):
            data = toml.loads(contents)
        else:
            data = json.loads(contents)

        return cls(data=_get_within(data, prefix), client_overrides=client_overrides)


def _get_within(
    data: dict,
    prefix: typing.Union[None, str, typing.Iterable[str]],
) -> dict:
    """
    Retrieves the lobotomy data within the root data object based on the given
    prefix. If there is no prefix, the root data object is the lobotomy data and
    that is returned instead.

    :param data:
        Loaded configuration data containing lobotomy data either at the top-level
        or as a descendent within the data dictionary as specified by the prefix.
    :param prefix:
        Optional prefix that specifies where within the data dictionary to find the
        lobotomized data. If not specified, the data dictionary is assumed to be a
        lobotomy data configuration dictionary. If specified, the prefix can either
        be represented as a hierarchical list of keys to descend within the root
        data dictionary, e.g. ["foo", "bar"] or as a dot-delimited string denoting
        the hierarchical list of keys, e.g. "foo.bar". If one of the keys contains
        a "." then the list form must be used.
    """
    if not prefix:
        return data

    parts = prefix.split(".") if isinstance(prefix, str) else prefix
    for key in parts:
        data = data.get(key, {})
    return data



import typing


class LobotomyError(Exception):
    """Base exception class for all lobotomy custom exceptions."""

    def __init__(self, message: str):
        super(LobotomyError, self).__init__(message)


class NoResponseFound(LobotomyError):
    """Error raised when no response was found for the given method call."""

    pass


class NoSuchMethod(LobotomyError):
    """Error raised when a service method definition cannot be found."""

    pass


class DataTypeError(LobotomyError):
    """Error raised during casting of data types that fails."""

    pass


class RequestValidationError(LobotomyError):
    """Error raised when validation of the request arguments fails for a method call."""

    pass


class ClientError(Exception):
    """
    An exception that mirrors the structure of a boto3/botocore exception
    without the added complexity of those actual exception classes.
    """

    def __init__(self, *args, **kwargs):
        super(ClientError, self).__init__()
        self._code: typing.Optional[str] = None
        self._message: typing.Optional[str] = None

    @property
    def response(self) -> typing.Dict[str, typing.Any]:
        return {"Error": {"Code": self._code, "Message": self._message}}

    def populate(self, **kwargs) -> "ClientError":
        self._code = kwargs.get("Code")
        self._message = kwargs.get("Message")
        return self



import argparse
import dataclasses
import typing


@dataclasses.dataclass(frozen=True)
class CliContext:
    """Data structure for CLI execution context."""

    #: Parsed arguments for the CLI invocation.
    args: argparse.Namespace


@dataclasses.dataclass(frozen=True)
class ExecutionResult:
    """Data structure for a CLI command execution result."""

    #: Exit identifier specifying how the command was resolved.
    code: str
    #: Message for the associated result.
    message: typing.Optional[str] = None

    def echo(self) -> "ExecutionResult":
        """Echoes the result if set for display to the console."""
        print(f"[{self.code}] {self.message}")
        return self



import typing

from lobotomy import _services


def add_service_response(
    lobotomy_data: dict,
    method: "_services.Method",
    response: typing.Any = None,
) -> None:
    if "clients" not in lobotomy_data:
        lobotomy_data["clients"] = {}

    if method.service.name not in lobotomy_data["clients"]:
        lobotomy_data["clients"][method.service.name] = {}

    service_data = lobotomy_data["clients"][method.service.name]
    response_type = method.output.get("type", "structure")
    if response is None:
        new_response = method.configuration_output
    else:
        new_response = response

    if method.name not in service_data:
        if response_type == "list":
            # List entries confuse the side effect behavior of popping
            # responses off the list. To avoid that the list responses
            # are added as list items.
            service_data[method.name] = [new_response]
        else:
            service_data[method.name] = new_response
        return

    existing = service_data[method.name]
    if isinstance(existing, dict):
        service_data[method.name] = [existing, new_response]
    else:
        service_data[method.name] = existing + [new_response]



import copy
import typing
import datetime


def parse_definition_item(shapes: dict, value: dict) -> dict:
    """
    Unwraps the value object by placing its shape data within a copy of
    the object itself and returning that copied object. This works
    recursively to unwrap results hierarchically.

    :param shapes:
        Shape definitions from the service spec in which this value
        definition resides.
    :param value:
        A definition object value to unwrap shape data within, or potentially
        a value object that has no shape and is already a leaf definition
        value as this function is called recursively.
    :return:
        An unwrapped, deep-copied version of the value dictionary where
        the shape(s) have been injected into the object hierarchically to
        remove the need for shape lookups to understand the properties,
        including types, of the value and potentially its member(s) if
        represents a list or dictionary.
    """
    output = copy.deepcopy(value)

    if value.get("type") == "map":
        # Maps are terminal but have key and value shapes.
        output["key"] = parse_definition_item(shapes, value["key"])
        output["value"] = parse_definition_item(shapes, value["value"])

    if "member" in output:
        # Types like lists have a member definition for the items in the
        # list. That needs to be parsed and then the shape is complete.
        output["member"] = parse_definition_item(shapes, output["member"])

    if "members" in output:
        output["members"] = {
            k: parse_definition_item(shapes, v)
            for k, v in (output["members"] or {}).items()
        }

    if "shape" in output:
        shape = copy.deepcopy(shapes[output["shape"]])
        output.update(parse_definition_item(shapes, shape))

    return output


def flat_cast(output: dict) -> typing.Any:
    """
    Creates a flat version of the output with default values for inclusion
    as a skeleton in a configuration response file.
    """
    o = output
    data_type = o.get("type", "structure")
    type_casts = {
        "blob": lambda: "...",
        "string": lambda: "...",
        "integer": lambda: 1,
        "long": lambda: 1,
        "float": lambda: 1.0,
        "timestamp": lambda: f'{datetime.datetime.utcnow().isoformat("T")}Z',
        "list": lambda: [flat_cast(o["member"])],
        # Some calls have empty requests or responses that have no members and so here
        # we allow for the empty possibility for the structure type.
        "structure": lambda: {k: flat_cast(v) for k, v in o.get("members", {}).items()},
    }
    caster = type_casts.get(data_type, lambda: None)
    return caster()
